# -*- coding: utf-8 -*-

import logging

from .query import Query
from .connection import ODataConnection


class Context:

    def __init__(self, session=None, auth=None):
        self.log = logging.getLogger('odata.context')
        self.connection = ODataConnection(session=session, auth=auth)

    def query(self, entitycls):
        q = Query(entitycls, connection=self.connection)
        return q

    def call(self, action_or_function, **parameters):
        """
        Call a defined Action or Function using this Context's connection

        :param action_or_function: Action/Function instance on a Entity class
        :param parameters: Keyword parameters to pass to Action/Function
        :return: OData raw response
        """
        parameters['__connection__'] = self.connection
        return action_or_function(**parameters)

    def delete(self, entity):
        """
        Creates a DELETE call to the service, deleting the entity

        :type entity: EntityBase
        :raises ODataConnectionError: Delete not allowed or a serverside error. Server returned an HTTP error code
        """
        self.log.info(u'Deleting entity: {0}'.format(entity))
        url = entity.__odata__.instance_url
        self.connection.execute_delete(url)
        entity.__odata__.persisted = False
        self.log.info(u'Success')

    def save(self, entity, force_refresh=True):
        """
        Creates a POST or PATCH call to the service. If the entity already has
        a primary key, an update is called. Otherwise the entity is inserted
        as new. Updating an entity will only send the changed values

        :param entity: Model instance to insert or update
        :type entity: EntityBase
        :param force_refresh: Read full entity data again from service after PATCH call
        :raises ODataConnectionError: Invalid data or serverside error. Server returned an HTTP error code
        """

        if self.is_entity_saved(entity):
            self._update_existing(entity, force_refresh=force_refresh)
        else:
            self._insert_new(entity)

    def is_entity_saved(self, entity):
        return entity.__odata__.persisted

    def _insert_new(self, entity):
        """
        Creates a POST call to the service, sending the complete new entity

        :type entity: EntityBase
        """
        self.log.info(u'Saving new entity')

        url = entity.__odata_url__()

        es = entity.__odata__
        insert_data = es.data_for_insert()
        saved_data = self.connection.execute_post(url, insert_data)
        es.reset()
        es.connection = self.connection
        es.persisted = True

        if saved_data is not None:
            es.update(saved_data)

        self.log.info(u'Success')

    def _update_existing(self, entity, force_refresh=True):
        """
        Creates a PATCH call to the service, sending only the modified values

        :type entity: EntityBase
        """
        es = entity.__odata__
        patch_data = es.data_for_update()

        if len([i for i in patch_data if not i.startswith('@')]) == 0:
            self.log.debug(u'Nothing to update: {0}'.format(entity))
            return

        self.log.info(u'Updating existing entity: {0}'.format(entity))

        url = es.instance_url

        saved_data = self.connection.execute_patch(url, patch_data)
        es.reset()

        if saved_data is None and force_refresh:
            self.log.info(u'Reloading entity from service')
            saved_data = self.connection.execute_get(url)

        if saved_data is not None:
            entity.__odata__.update(saved_data)

        self.log.info(u'Success')



import os
import os.path
import json

class Settings:
    Apikey = ""
    Scheme = "http"
    Host = "127.0.0.1"
    Port = 8181
    UseCloudFX = False
    def __init__(self):
        # import os.path
        fn = os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir))
        fn = os.path.join(fn, 'settings.json')
        # TODO: If 'settings.json' does not exist then create with default values and tell user to edit the file
        # Alternatively prompt for APIKey etc.
        # fn = os.path.join(os.path.dirname(__file__), 'settings.json')
        if not os.path.isfile(fn):
            with open(fn, 'w') as outfile:
                json.dump({
                    'Apikey':self.Apikey,
                    'Scheme':self.Scheme,
                    'Host':self.Host,
                    'Port':self.Port,
                    'UseCloudFX':self.UseCloudFX
                    }, outfile, indent=4, sort_keys=True)
        else:
            json_data=open(fn).read()
            data = json.loads(json_data)
            self.__dict__.update(data)

    def getUrl(self, path):
        if self.UseCloudFX:
            return "https://api.unifiedfx.com/" + path
        else:
            url = "{0}://{1}:{2}/AutomationFX/api/{3}"
            url = url.format(self.Scheme,self.Host,self.Port, path)
            return url




# -*- coding: utf-8 -*-

"""
Exceptions
==========

Error classes used by this library.
"""


class ODataError(Exception):
    """
    Base class for python-odata errors. All other errors are subclassed
    from this. Raising any other exception class is a bug and should be
    reported
    """
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)

        self.status_code = None
        """HTTP status"""

        self.code = None
        """Error code supplied by server"""

        self.message = None
        """Error message supplied by server"""

        self.detailed_message = None
        """Detailed error message supplied by server"""


class ODataConnectionError(ODataError):
    """Raised when the endpoint responds with an HTTP error code"""
    pass


class ODataQueryError(ODataError):
    pass


class ODataReflectionError(ODataError):
    """
    Raised when MetaData is unable to reflect types
    """
    pass


class NoResultsFound(ODataQueryError):
    """
    Raised when :py:func:`~odata.query.Query.one` is called but zero results
    returned
    """
    pass


class MultipleResultsFound(ODataQueryError):
    """
    Raised when :py:func:`~odata.query.Query.one` is called but multiple results
    returned
    """
    pass



# -*- coding: utf-8 -*-

"""
Actions, Functions
==================

Actions and Functions are set up automatically when using reflection. Unbound
callables are collected in :py:attr:`~odata.service.ODataService.actions` and
:py:attr:`~odata.service.ODataService.functions`. Bound callables are assigned
to the Entity classes they are bound to.

.. code-block:: python

    >>> from odata import ODataService
    >>> Service = ODataService(url, reflect_entities=True)
    >>> Product = Service.entities['Product']

    >>> prod = Service.query(Product).get(1234)
    >>> prod.GetAvailabilityDate()
    datetime.datetime(2018, 6, 1, 12, 0, 0)

    >>> import datetime
    >>> GetExampleDecimal = Service.functions['GetExampleDecimal']
    >>> GetExampleDecimal(Date=datetime.datetime.now())
    Decimal('34.0')


Unbound action/functions
------------------------

Similar to Entities, Functions are subclassed from the Service baseclasses
:py:attr:`~odata.service.ODataService.Action` and
:py:attr:`~odata.service.ODataService.Function`:

.. code-block:: python

    Service = ODataService(url)

    class _GetExampleDecimal(Service.Function):
        name = 'GetExampleDecimal'
        parameters = dict(
            Date=DatetimeProperty,
        )
        return_type = DecimalProperty

    GetExampleDecimal = _GetExampleDecimal()

Usage:

.. code-block:: python

    >>> import datetime
    >>> # calls GET http://service/GetExampleDecimal(Date=2017-01-01T12:00:00Z)
    >>> GetExampleDecimal(Date=datetime.datetime.now())
    Decimal('34.0')


Bound action/function
---------------------

Bound functions are otherwise the same, but the instanced object should be set
under the Entity it belongs to:

.. code-block:: python

    class _GetAvailabilityDate(Service.Function):
        name = 'ODataService.GetAvailabilityDate'
        parameters = dict()
        return_type = DatetimeProperty

    class _RemoveAllReservations(Service.Action):
        name = 'ODataService.RemoveAllReservations'
        parameters = dict()
        return_type = BooleanProperty

    class _ReserveAmount(Service.Action):
        name = 'ODataService.ReserveAmount'
        parameters = dict(
            Amount=DecimalProperty,
        )
        return_type = BooleanProperty

    class Product(Service.Entity):
        Id = IntegerProperty('Id', primary_key=True)
        Name = StringProperty('Name')

        GetAvailabilityDate = GetAvailabilityDate()
        RemoveAllReservations = _RemoveAllReservations()
        ReserveAmount = _ReserveAmount()

Usage:

.. code-block:: python

    >>> # collection bound Action. calls POST http://service/Product/ODataService.RemoveAllReservations
    >>> Product.RemoveAllReservations()
    True

    >>> # if the Action is instance bound, call the Action from the Product instance instead
    >>> from decimal import Decimal
    >>> prod = Service.query(Product).get(1234)
    >>> # calls POST http://service/Product(1234)/ODataService.ReserveAmount
    >>> prod.ReserveAmount(Amount=Decimal('5.0'))
    True

    >>> # calls GET http://service/Product(1234)/ODataService.GetAvailabilityDate()
    >>> prod.GetAvailabilityDate()
    datetime.datetime(2018, 6, 1, 12, 0, 0)


----

API
---
"""

try:
    # noinspection PyUnresolvedReferences
    from urllib.parse import urljoin
except ImportError:
    # noinspection PyUnresolvedReferences
    from urlparse import urljoin
from collections import OrderedDict


class ActionCallable(object):
    """
    A helper class for ActionBase, representing a callable
    """
    def __init__(self, actionbase_instance, url, errmsg=None):
        self.actionbase_instance = actionbase_instance
        self.url = url
        self.errmsg = errmsg
        self.query = None

    def __repr__(self):
        name = self.actionbase_instance.name
        url = self.url + '/' + name
        return '<Callable for {0}>'.format(url)

    def with_query(self, query):
        self.query = query
        return self

    def __call__(self, **kwargs):
        if self.errmsg is not None:
            raise AttributeError(self.errmsg)

        connection = self.actionbase_instance._get_context_or_default_connection(kwargs)
        return self.actionbase_instance._callable(connection, self.url, self.query, **kwargs)


class ActionBase(object):

    __odata_service__ = None

    name = None
    """
    Action's fully qualified name (usually SchemaName.ActionName)

    :type name: str
    """

    parameters = None
    """
    A dictionary that defines what keyword arguments and what types this Action
    accepts. For example, ``dict(productId=IntegerProperty)``

    :type parameters: dict
    """

    return_type_collection = None
    """
    Reference to the returned value's type, when retuning multiple values.
    Entity or Property class
    """
    return_type = None
    """
    Reference to the returned value's type. Entity or Property class
    """

    bound_to_collection = False
    """
    Action is bound to Entity collection. If True, Action is not available in
    Entity class instances. If False, Action is not available in Entity class
    objects
    """

    def __get__(self, instance, owner):
        # return a callable that acts on EntitySet or the Entity itself

        errmsg = None
        if self.bound_to_collection and instance is not None:
            errmsg = '{0} is bound to collection {1}'.format(self.name, owner)
        if self.bound_to_collection is False:
            if instance is None:
                errmsg = '{0} is bound to instances of {1}'.format(self.name, owner)
            else:
                es = instance.__odata__
                if es.persisted is False:
                    errmsg = ('Trying to call instance bound Action or '
                              'Function \'{0}\', but instance {1} is not saved'
                              ''.format(self.name, instance))

        url = None
        if instance:
            # /MyEntity(1234)/SchemaName.ActionName
            url = instance.__odata__.instance_url

        # default to /MyEntity/SchemaName.ActionName
        url = url or owner.__odata_url__()

        ac = ActionCallable(self, url, errmsg=errmsg)
        return ac

    def __call__(self, **kwargs):
        url = self.__odata_service__.url
        ac = ActionCallable(self, url)
        # call immediately, otherwise user needs to double-call. this removes
        # the ability to query results
        return ac(**kwargs)

    def _get_context_or_default_connection(self, kwargs):
        connection = kwargs.pop('__connection__', None)
        connection = connection or self.__odata_service__.default_context.connection
        return connection

    def _check_call_arguments(self, kwargs):
        incorrect_keys = set(kwargs.keys()) != set(self.parameters.keys())
        if incorrect_keys:
            received_keys = ','.join(kwargs.keys())
            expected_keys = ','.join(self.parameters.keys())
            errmsg = 'Received keyword arguments: \'{}\', required: \'{}\''
            errmsg = errmsg.format(received_keys, expected_keys)
            raise TypeError(errmsg)

    def _callable(self, connection, url, query, **kwargs):
        self._check_call_arguments(kwargs)

        if not url.endswith('/'):
            url += '/'
        url += self.name

        query_options = None
        if query:
            query_options = query._get_options()

        response_data = self._execute_http(connection, url, query_options, kwargs)
        response_data = (response_data or {}).get('value')

        simple_types_values = self.__odata_service__.metadata.property_types.values()

        if self.return_type_collection:
            if self.return_type_collection in simple_types_values:
                values_collection = []
                prop = self.return_type_collection
                for value in response_data:
                    deserialized = prop('temp').deserialize(value)
                    values_collection.append(deserialized)
                return values_collection

            entity_collection = []
            for value in (response_data or []):
                entity_instance = self.return_type_collection.__new__(self.return_type_collection, from_data=value)
                entity_collection.append(entity_instance)
            return entity_collection

        if self.return_type:
            if self.return_type in simple_types_values:
                prop = self.return_type
                return prop('temp').deserialize(response_data)

            entity_instance = self.return_type.__new__(self.return_type, from_data=response_data)
            return entity_instance

        # no defined type, return whatever we got
        return response_data

    def _execute_http(self, connection, url, query_options, kwargs):
        raise NotImplementedError()


class Action(ActionBase):
    """
    Baseclass for all Actions. Should not be used directly, use the
    subclass :py:class:`~odata.service.ODataService.Action` instead.

    .. py:attribute:: name

        Action's fully qualified name. Bound Actions are prefixed with their
        schema's name (``SchemaName.ActionName``).
        **Required when subclassing**

    .. py:attribute:: parameters

        Dictionary. Defines all the keyword arguments and their types the
        Action can accept. Key names must match with ones accepted by the
        server.
        **Required when subclassing**

    .. py:attribute:: return_type

        Reference to either Entity class or Property class. Defines the return
        value's type for this Action.
        **Required when subclassing**

    .. py:attribute:: return_type_collection

        Reference to either Entity class or Property class. Defines the return
        value's type for this Action when retuning multiple values.
        **Required when subclassing**

    .. py:attribute:: bound_to_collection

        Action is bound to Entity collection. If True, Action is not available
        in Entity class instances. If False, Action is not available in Entity
        class objects. Defaults to *false*
    """

    name = 'ODataSchema.Action'

    def _execute_http(self, connection, url, query_options, kwargs):
        # Execute http POST, encoding kwargs to json body
        data = OrderedDict()
        for key, value in kwargs.items():
            prop_type = self.parameters.get(key)
            escaped_value = prop_type('temp').serialize(value)
            data[key] = escaped_value

        return connection.execute_post(url, data, params=query_options)


class Function(ActionBase):
    """
    Baseclass for all Functions. Should not be used directly, use the
    subclass :py:class:`~odata.service.ODataService.Function` instead.

    .. py:attribute:: name

        Function's fully qualified name. Function Actions are prefixed with
        their schema's name (``SchemaName.FunctionName``).
        **Required when subclassing**

    .. py:attribute:: parameters

        Dictionary. Defines all the keyword arguments and their types the
        Function can accept. Key names must match with ones accepted by the
        server.
        **Required when subclassing**

    .. py:attribute:: return_type

        Reference to either Entity class or Property class. Defines the return
        value's type for this Function.
        **Required when subclassing**

    .. py:attribute:: return_type_collection

        Reference to either Entity class or Property class. Defines the return
        value's type for this Function when retuning multiple values.
        **Required when subclassing**

    .. py:attribute:: bound_to_collection

        Function is bound to Entity collection. If True, Function is not
        available in Entity class instances. If False, Function is not
        available in Entity class objects. Defaults to *false*
    """

    name = 'ODataSchema.Function'

    def _execute_http(self, connection, url, query_options, kwargs):
        # Execute http GET, passing kwargs as parameters in url
        kwargs_escaped = []
        for key, value in kwargs.items():
            prop_type = self.parameters.get(key)
            escaped_value = prop_type('temp').escape_value(value)
            kwargs_escaped.append((key, escaped_value))

        kwargs_escaped = sorted(kwargs_escaped, key=lambda x: x[0])
        params = ['='.join([key, str(value)]) for key, value in kwargs_escaped]
        params = ','.join(params)
        url += '({0})'.format(params)

        return connection.execute_get(url, params=query_options)



# -*- coding: utf-8 -*-

from odata import ODataService
from odata.property import StringProperty, IntegerProperty, DecimalProperty, \
    NavigationProperty, DatetimeProperty
from odata.enumtype import EnumType, EnumTypeProperty

url = 'http://unittest.server.local/odata/'
Service = ODataService(url)


class DemoActionWithParameters(Service.Action):
    name = 'ODataTest.DemoActionParameters'
    parameters = dict(
        Name=StringProperty,
        Price=DecimalProperty,
    )
    bound_to_collection = True


class DemoAction(Service.Action):
    name = 'ODataTest.DemoAction'
    parameters = {}


class DemoCollectionAction(Service.Action):
    name = 'ODataTest.DemoCollectionAction'
    parameters = {}
    bound_to_collection = True


class _DemoUnboundAction(Service.Action):
    name = 'ODataTest.DemoUnboundAction'
    parameters = {}

DemoUnboundAction = _DemoUnboundAction()


class DemoFunction(Service.Function):
    name = 'ODataTest.DemoFunction'
    parameters = {}
    bound_to_collection = True


class ColorSelection(EnumType):
    Black = 0
    Red = 1
    Blue = 2
    Green = 3


class Product(Service.Entity):
    __odata_type__ = 'ODataTest.Objects.Product'
    __odata_collection__ = 'ProductParts'

    id = IntegerProperty('ProductID', primary_key=True)
    name = StringProperty('ProductName')
    category = StringProperty('Category')
    price = DecimalProperty('Price')
    color_selection = EnumTypeProperty('ColorSelection',
                                       enum_class=ColorSelection)

    DemoAction = DemoAction()
    DemoCollectionAction = DemoCollectionAction()
    DemoActionWithParameters = DemoActionWithParameters()
    DemoFunction = DemoFunction()


class ProductPart(Service.Entity):
    __odata_type__ = 'ODataTest.Objects.ProductPart'
    __odata_collection__ = 'ProductParts'
    id = IntegerProperty('PartID', primary_key=True)
    name = StringProperty('PartName')
    size = DecimalProperty('Size')
    product_id = IntegerProperty('ProductID')


class Manufacturer(Service.Entity):
    __odata_type__ = 'ODataTest.Objects.Manufacturer'
    __odata_collection__ = 'Manufacturers'

    id = IntegerProperty('ManufacturerID', primary_key=True)
    name = StringProperty('Name')
    established_date = DatetimeProperty('DateEstablished')


class ProductWithNavigation(Product):
    __odata_type__ = 'ODataTest.Objects.ProductWithNavigation'
    __odata_collection__ = 'ProductsWithNavigation'

    manufacturer_id = IntegerProperty('ManufacturerID')

    manufacturer = NavigationProperty('Manufacturer', Manufacturer, foreign_key=manufacturer_id)
    parts = NavigationProperty('Parts', ProductPart, collection=True)

ProductPart.product = NavigationProperty('Product', ProductWithNavigation, foreign_key=ProductPart.product_id)


class ProductManufacturerSales(Service.Entity):
    __odata_type__ = 'ODataTest.Objects.ProductManufacturerSales'
    __odata_collection__ = 'Product_Manufacturer_Sales'

    product_id = IntegerProperty('ProductID', primary_key=True)
    manufacturer_id = IntegerProperty('ManufacturerID', primary_key=True)
    sales_amount = DecimalProperty('SalesAmount')



# -*- coding: utf-8 -*-

"""
Connecting to an endpoint
=========================

A service object represents a single endpoint which has at least one EntitySet.
It is the cornerstone of this library. If you have multiple endpoints to
connect to, create multiple instances of :py:class:`ODataService`. All Entity
objects are each bound to exactly one Service and cannot be used across
multiple services.

Optionally, the Service object can connect to the endpoint and request its
metadata document. This document will then be used to build Entity objects
corresponding to each EntitySet provided by the endpoint. This operation
requires a working network connection to the endpoint. Creating an instance with
``reflect_entities=False`` will not cause any network activity.


Authentication
--------------

``auth`` and ``session`` keyword arguments to :py:class:`ODataService` are
passed as-is to Requests calls, so most of the same `guides`_ can be used.

.. _guides: http://docs.python-requests.org/en/master/user/authentication/


HTTP Basic authentication:

.. code-block:: python

    >>> from requests.auth import HTTPBasicAuth
    >>> my_auth = HTTPBasicAuth('username', 'password')
    >>> Service = ODataService('url', auth=my_auth)


NTLM Auth (for services like Microsoft Dynamics 2016):

.. code-block:: python

    >>> import requests
    >>> from requests_ntlm import HttpNtlmAuth
    >>> my_session = requests.Session()
    >>> my_session.auth = HttpNtlmAuth('domain\\username', 'password')
    >>> my_session.get('basic url')  # should return 200 OK
    >>> Service = ODataService('url', session=my_session)


----

API
---
"""

import logging

from .entity import EntityBase, declarative_base
from .metadata import MetaData
from .exceptions import ODataError
from .context import Context
from .action import Action, Function

__all__ = (
    'ODataService',
    'ODataError',
)


class ODataService(object):
    """
    :param url: Endpoint address. Must be an address that can be appended with ``$metadata``
    :param base: Custom base class to use for entities
    :param reflect_entities: Create a request to the service for its metadata, and create entity classes automatically
    :param session: Custom Requests session to use for communication with the endpoint
    :param auth: Custom Requests auth object to use for credentials
    :raises ODataConnectionError: Fetching metadata failed. Server returned an HTTP error code
    """
    def __init__(self, url, base=None, reflect_entities=False, session=None, auth=None):
        self.url = url
        self.metadata_url = ''
        self.collections = {}
        self.log = logging.getLogger('odata.service')
        self.default_context = Context(auth=auth, session=session)

        self.entities = {}
        """
        A dictionary containing all the automatically created Entity classes.
        Empty if the service is created with ``reflect_entities=False``

        :type entities: dict
        """
        self.actions = {}
        """
        A dictionary containing all the automatically created unbound Action
        callables. Empty if the service is created with
        ``reflect_entities=False``

        :type actions: dict
        """
        self.functions = {}
        """
        A dictionary containing all the automatically created unbound Function
        callables. Empty if the service is created with
        ``reflect_entities=False``

        :type functions: dict
        """
        self.types = {}
        """
        A dictionary containing all types (EntityType, EnumType) created
        during reflection. Empty if the service is created with
        ``reflect_entities=False``

        :type types: dict
        """

        self.metadata = MetaData(self)
        self.Base = base or declarative_base()
        """
        Entity base class. Either a custom one given in init or a generated one. Can be used to define entities

        :type Base: EntityBase
        """
        self.Entity = self.Base  # alias

        self.Action = type('Action', (Action,), dict(__odata_service__=self))
        """
        A baseclass for this service's Actions

        :type Action: Action
        """

        self.Function = type('Function', (Function,), dict(__odata_service__=self))
        """
        A baseclass for this service's Functions

        :type Function: Function
        """

        if reflect_entities:
            _, self.entities, self.types = self.metadata.get_entity_sets(base=self.Entity)

        self.Entity.__odata_url_base__ = url
        self.Entity.__odata_service__ = self

    def __repr__(self):
        return u'<ODataService at {0}>'.format(self.url)

    def create_context(self, auth=None, session=None):
        """
        Create new context to use for session-like usage

        :param auth: Custom Requests auth object to use for credentials
        :param session: Custom Requests session to use for communication with the endpoint
        :return: Context instance
        :rtype: Context
        """
        return Context(auth=auth, session=session)

    def describe(self, entity):
        """
        Print a debug screen of an entity instance

        :param entity: Entity instance to describe
        """
        entity.__odata__.describe()

    def is_entity_saved(self, entity):
        """Returns boolean indicating entity's status"""
        return self.default_context.is_entity_saved(entity)

    def query(self, entitycls):
        """
        Start a new query for given entity class

        :param entitycls: Entity to query
        :return: Query object
        """
        return self.default_context.query(entitycls)

    def delete(self, entity):
        """
        Creates a DELETE call to the service, deleting the entity

        :type entity: EntityBase
        :raises ODataConnectionError: Delete not allowed or a serverside error. Server returned an HTTP error code
        """
        return self.default_context.delete(entity)

    def save(self, entity, force_refresh=True):
        """
        Creates a POST or PATCH call to the service. If the entity already has
        a primary key, an update is called. Otherwise the entity is inserted
        as new. Updating an entity will only send the changed values

        :param entity: Model instance to insert or update
        :param force_refresh: Read full entity data again from service after PATCH call
        :raises ODataConnectionError: Invalid data or serverside error. Server returned an HTTP error code
        """
        return self.default_context.save(entity, force_refresh=force_refresh)



# -*- coding: utf-8 -*-

"""
Querying
========

Entities can be queried from a service with a Query object:

.. code-block:: python

    query = Service.query(Order)

Adding filters and other options always creates a new Query object with the
given directives:

.. code-block:: python

    >>> query.filter(Order.Name == 'Foo')
    <Query for <Order>>

This makes object chaining possible:

.. code-block:: python

    >>> first_order = query.filter(...).filter(...).order_by(...).first()

The resulting objects can be fetched with :py:func:`~Query.first`,
:py:func:`~Query.one`, :py:func:`~Query.all`, :py:func:`~Query.get` or
just iterating the Query object itself. Network is not accessed until one of
these ways is triggered.

Navigation properties can be loaded in the same request with 
:py:func:`~Query.expand`:

.. code-block:: python

    >>> query.expand(Order.Shipper, Order.Customer)
    >>> order = query.first()

----

API
---
"""

try:
    # noinspection PyUnresolvedReferences
    from urllib.parse import urljoin
except ImportError:
    # noinspection PyUnresolvedReferences
    from urlparse import urljoin

from .exceptions import *

# from .exceptions import * as exc

# import exceptions as exc


class Query(object):
    """
    This class should not be instantiated directly, but from a
    :py:class:`~odata.service.ODataService` object.
    """
    def __init__(self, entitycls, connection=None, options=None):
        self.entity = entitycls
        self.options = options or dict()
        self.connection = connection

    def __iter__(self):
        url = self._get_url()
        options = self._get_options()
        while True:
            data = self.connection.execute_get(url, options)
            if 'value' in data:
                value = data.get('value', [])
                for row in value:
                    yield self._create_model(row)

                if '@odata.nextLink' in data:
                    url = urljoin(self.entity.__odata_url_base__, data['@odata.nextLink'])
                    options = {}  # we get all options in the nextLink url
                else:
                    break

    def __repr__(self):
        return '<Query for {0}>'.format(self.entity)

    def __str__(self):
        return self.as_string()

    def _get_url(self):
        return self.entity.__odata_url__()

    def _get_options(self):
        """
        Format current query options to a dict that can be passed to requests
        :return: Dictionary
        """
        options = dict()

        _top = self.options.get('$top')
        if _top is not None:
            options['$top'] = _top

        _offset = self.options.get('$skip')
        if _offset is not None:
            options['$skip'] = _offset

        _select = self.options.get('$select')
        if _select:
            options['$select'] = ','.join(_select)

        _filters = self.options.get('$filter')
        if _filters:
            options['$filter'] = ' and '.join(_filters)

        _expand = self.options.get('$expand')
        if _expand:
            options['$expand'] = ','.join(_expand)

        _order_by = self.options.get('$orderby')
        if _order_by:
            options['$orderby'] = ','.join(_order_by)
        return options

    def _create_model(self, row):
        if len(self.options.get('$select', [])):
            return row
        else:
            e = self.entity.__new__(self.entity, from_data=row)
            es = e.__odata__
            es.connection = self.connection
            return e

    def _get_or_create_option(self, name):
        if name not in self.options:
            self.options[name] = []
        return self.options[name]

    def _format_params(self, options):
        return '&'.join(['='.join((key, str(value))) for key, value in options.items() if value is not None])

    def _new_query(self):
        """
        Create copy of this query without mutable values. All query builders
        should use this first.

        :return: Query instance
        """
        o = dict()
        o['$top'] = self.options.get('$top', None)
        o['$skip'] = self.options.get('$skip', None)
        o['$select'] = self.options.get('$select', [])[:]
        o['$filter'] = self.options.get('$filter', [])[:]
        o['$expand'] = self.options.get('$expand', [])[:]
        o['$orderby'] = self.options.get('$orderby', [])[:]
        return Query(self.entity, options=o, connection=self.connection)

    def as_string(self):
        query = self._format_params(self._get_options())
        return urljoin(self._get_url(), '?{0}'.format(query))

    # Query builders ###########################################################

    def select(self, *values):
        """
        Set properties to fetch instead of full Entity objects

        :return: Raw JSON values for given properties
        """
        q = self._new_query()
        option = q._get_or_create_option('$select')
        for prop in values:
            option.append(prop.name)
        return q

    def filter(self, value):
        """
        Set ``$filter`` query parameter. Can be called multiple times. Multiple
        :py:func:`filter` calls are concatenated with 'and'

        :param value: Property comparison. For example, ``Entity.Property == 2``
        :return: Query instance
        """
        q = self._new_query()
        option = q._get_or_create_option('$filter')
        option.append(value)
        return q

    def expand(self, *values):
        """
        Set ``$expand`` query parameter

        :param values: ``Entity.Property`` instance
        :return: Query instance
        """
        q = self._new_query()
        option = q._get_or_create_option('$expand')
        for prop in values:
            option.append(prop.name)
        return q

    def order_by(self, *values):
        """
        Set ``$orderby`` query parameter

        :param values: One of more of Property.asc() or Property.desc()
        :return: Query instance
        """
        q = self._new_query()
        option = q._get_or_create_option('$orderby')
        option.extend(values)
        return q

    def limit(self, value):
        """
        Set ``$top`` query parameter

        :param value: Number of records to return
        :return: Query instance
        """
        q = self._new_query()
        q.options['$top'] = value
        return q

    def offset(self, value):
        """
        Set ``$skip`` query parameter

        :param value: Number of records to skip
        :return: Query instance
        """
        q = self._new_query()
        q.options['$skip'] = value
        return q

    @staticmethod
    def and_(value1, value2):
        return '{0} and {1}'.format(value1, value2)

    @staticmethod
    def or_(value1, value2):
        return '{0} or {1}'.format(value1, value2)

    @staticmethod
    def grouped(value):
        return '({0})'.format(value)

    # Actions ##################################################################

    def all(self):
        """
        Returns a list of all Entity instances that match the current query
        options. Iterates through all results with multiple requests fired if
        necessary, exhausting the query

        :return: A list of Entity instances
        """
        return list(iter(self))

    def first(self):
        """
        Return the first Entity instance that matches current query

        :return: Entity instance or None
        """
        oldvalue = self.options.get('$top', None)
        self.options['$top'] = 1
        data = list(iter(self))
        self.options['$top'] = oldvalue
        if data:
            return data[0]

    def one(self):
        """
        Return only one resulting Entity

        :return: Entity instance
        :raises NoResultsFound: Zero results returned
        :raises MultipleResultsFound: Multiple results returned
        """
        oldlimit = self.options.get('$top', None)

        self.options['$top'] = 2
        data = self.all()

        self.options['$top'] = oldlimit
        if len(data) == 0:
            raise NoResultsFound()
        if len(data) > 1:
            raise MultipleResultsFound()
        return data[0]

    def get(self, *pk, **composite_keys):
        """
        Return a Entity with the given primary key

        :param pk: Primary key value
        :param composite_keys: Primary key values for Entities with composite keys
        :return: Entity instance or None
        """
        i = self.entity.__new__(self.entity)
        es = i.__odata__

        oldfilters = self._get_or_create_option('$filter')

        tempfilters = []

        if pk:
            pk = pk[0]
            prop = es.primary_key_properties[0][1]
            tempfilters.append(prop == pk)
        else:
            for _, prop in es.primary_key_properties:
                tempfilters.append(prop == composite_keys[prop.name])

        self.options['$filter'] = tempfilters
        data = list(iter(self))

        self.options['$filter'] = oldfilters
        if len(data) > 0:
            return data[0]
        raise NoResultsFound()

    def raw(self, query_params):
        """
        Execute a query with custom parameters. Allows queries that
        :py:class:`Query` does not support otherwise. Results are not converted
        to Entity objects

        .. code-block:: python

            >>> query = Service.query(MyEntity)
            >>> query.raw({'$filter': 'EntityId eq 123456'})
            [{'EntityId': 123456, 'Name': 'Example entity'}]

        :param query_params: A dictionary of query params containing $filter, $orderby, etc.
        :type query_params: dict
        :return: Query result
        """
        url = self.entity.__odata_url__()
        response_data = self.connection.execute_get(url, params=query_params)
        return (response_data or {}).get('value')



# -*- coding: utf-8 -*-

"""
Entity classes
==============

The data model can be created manually if you wish to use separate property
names from the data keys, or define custom methods for your objects.

Custom Entity class
-------------------

Each Service instance has their separate base classes for Entities,
Actions and Functions. Use them to define your model:

.. code-block:: python

    class Product(Service.Entity):
        __odata_type__ = 'ProductDataService.Objects.Product'
        __odata_collection__ = 'Products'

        name = StringProperty('ProductName')
        quantity_in_storage = IntegerProperty('QuantityInStorage')

Note that the type (EntityType) and collection (EntitySet) must be defined.
These are used in querying and saving data.


Custom base class
-----------------

Define a base. These properties and methods are shared by all objects in the endpoint.

.. code-block:: python

    from odata.entity import declarative_base
    from odata.property import IntegerProperty, StringProperty, DatetimeProperty

    class MyBase(declaractive_base()):
        id = IntegerProperty('Id', primary_key=True)
        created_date = DatetimeProperty('Created')
        modified_date = DatetimeProperty('Modified')

        def did_somebody_touch_this(self):
            return self.created_date != self.modified_date

Define a model:

.. code-block:: python

    class Product(MyBase):
        __odata_type__ = 'ProductDataService.Objects.Product'
        __odata_collection__ = 'Products'

        name = StringProperty('ProductName')
        quantity_in_storage = IntegerProperty('QuantityInStorage')

        def is_product_available(self):
            return self.quantity_in_storage > 0


Use the base to init :py:class:`~odata.service.ODataService`:

.. code-block:: python

    Service = ODataService(url, base=MyBase)

Unlike reflection, this does not require any network connections. Now you can
use the Product class to create new objects or query existing ones:

.. code-block:: python

    query = Service.query(Product)
    query = query.filter(Product.name.startswith('Kettle'))
    for product in query:
        print(product.name, product.is_product_available())
"""

try:
    # noinspection PyUnresolvedReferences
    from urllib.parse import urljoin
except ImportError:
    # noinspection PyUnresolvedReferences
    from urlparse import urljoin

from .state import EntityState


class EntityBase(object):
    __odata_service__ = None
    __odata_collection__ = None
    __odata_type__ = 'ODataSchema.Entity'

    @classmethod
    def __odata_url__(cls):
        # used by Query
        if cls.__odata_collection__:
            return urljoin(cls.__odata_service__.url, cls.__odata_collection__)

    def __new__(cls, *args, **kwargs):
        i = super(EntityBase, cls).__new__(cls)
        i.__odata__ = es = EntityState(i)

        if 'from_data' in kwargs:
            raw_data = kwargs.pop('from_data')

            # check for values from $expand
            for prop_name, prop in es.navigation_properties:
                if prop.name in raw_data:
                    expanded_data = raw_data.pop(prop.name)
                    if prop.is_collection:
                        es.nav_cache[prop.name] = dict(collection=prop.instances_from_data(expanded_data))
                    else:
                        es.nav_cache[prop.name] = dict(single=prop.instances_from_data(expanded_data))

            for prop_name, prop in es.properties:
                i.__odata__[prop.name] = raw_data.get(prop.name)

            i.__odata__.persisted = True
        else:
            for prop_name, prop in es.properties:
                i.__odata__[prop.name] = None

        return i

    def __repr__(self):
        clsname = self.__class__.__name__
        display_string = self.__odata__.id or clsname
        return '<Entity({0})>'.format(display_string)

    def __eq__(self, other):
        if isinstance(other, EntityBase):
            my_id = self.__odata__.id
            if my_id:
                return my_id == other.__odata__.id
        return False


def declarative_base():
    return type('Entity', (EntityBase,), dict())



# -*- coding: utf-8 -*-

import unittest
import json
from decimal import Decimal

import responses
import requests

from odata.tests import Service, Product, ProductWithNavigation, ProductPart


class TestSimpleObjectManipulation(unittest.TestCase):

    def test_create(self):
        # Post call ###########################################################
        def request_callback(request):
            payload = json.loads(request.body)

            assert 'OData-Version' in request.headers, 'OData-Version header not in request'

            assert 'ProductID' not in payload, 'Payload contains primary key'
            assert '@odata.type' in payload, 'Payload did not contain @odata.type'

            payload['ProductID'] = 1

            resp_body = payload
            headers = {}
            return requests.codes.created, headers, json.dumps(resp_body)

        new_product = Product()
        new_product.name = u'New Test Product'
        new_product.category = u'Category #1'
        new_product.price = 34.5

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                rsps.POST, Product.__odata_url__(),
                callback=request_callback,
                content_type='application/json',
            )

            Service.save(new_product)

        assert new_product.id is not None, 'Product.id is not set'

    def test_create_with_primary_key(self):
        # Post call ###########################################################
        def request_callback(request):
            payload = json.loads(request.body)
            self.assertEqual(payload.get('ProductID'), 55,
                             msg='Did not receive ProductID')
            resp_body = payload
            headers = {}
            return requests.codes.created, headers, json.dumps(resp_body)

        new_product = Product()
        new_product.id = 55
        new_product.name = u'New Test Product'
        new_product.category = u'Category #1'
        new_product.price = 34.5

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                rsps.POST, Product.__odata_url__(),
                callback=request_callback,
                content_type='application/json',
            )

            Service.save(new_product)

        assert new_product.id is not None, 'Product.id is not set'

    def test_create_deep_inserts(self):
        # Initial part data ###################################################
        def request_callback_part(request):

            payload = {
                'PartID': 35,
                'PartName': 'Testing',
                'Size': 55,
            }

            resp_body = {'value': [payload]}
            headers = {}
            return requests.codes.created, headers, json.dumps(resp_body)

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                rsps.GET, ProductPart.__odata_url__(),
                callback=request_callback_part,
                content_type='application/json',
            )

            queried_part = Service.query(ProductPart).first()

        # Post call ###########################################################
        def request_callback(request):
            payload = json.loads(request.body)

            assert 'OData-Version' in request.headers, 'OData-Version header not in request'

            assert 'ProductID' not in payload, 'Payload contains primary key'
            assert '@odata.type' in payload, 'Payload did not contain @odata.type'

            assert 'Parts@odata.bind' in payload, 'Parts bind not in payload'
            assert payload['Parts@odata.bind'] == ['ProductParts(35)']

            assert 'Parts' in payload, 'Parts not in payload'
            parts = payload['Parts']
            assert isinstance(parts, list), 'Parts is not a list'
            part = parts[0]
            assert len(part) == 3, 'Extra keys in deep inserted part: {0}'.format(part)
            assert '@odata.type' in part
            assert part['PartName'] == 'Foo'
            assert part['Size'] == 12.5

            payload['ProductID'] = 1

            resp_body = payload
            headers = {}
            return requests.codes.created, headers, json.dumps(resp_body)

        part = ProductPart()
        part.name = 'Foo'
        part.size = 12.5

        new_product = ProductWithNavigation()
        new_product.name = u'New Test Product'
        new_product.category = u'Category #1'
        new_product.price = 34.5
        new_product.parts = [part, queried_part]

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                rsps.POST, ProductWithNavigation.__odata_url__(),
                callback=request_callback,
                content_type='application/json',
            )

            Service.save(new_product)

        assert new_product.id is not None, 'Product.id is not set'

    def test_read(self):
        expected_id = 1024
        expected_name = 'Existing entity'
        expected_category = 'Existing category'
        expected_price = Decimal('85.2')

        # Initial data ########################################################
        def request_callback(request):
            payload = {
                'ProductID': expected_id,
                'ProductName': expected_name,
                'Category': expected_category,
                'Price': float(expected_price),
            }

            resp_body = {'value': [payload]}
            headers = {}
            return requests.codes.ok, headers, json.dumps(resp_body)

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                rsps.GET, Product.__odata_url__(),
                callback=request_callback,
                content_type='application/json',
            )

            product = Service.query(Product).first()

        assert product.id == expected_id
        assert product.name == expected_name
        assert product.category == expected_category
        assert product.price == expected_price

    def test_update(self):
        expected_id = 1024
        expected_name = 'Existing entity'
        expected_category = 'Existing category'
        expected_price = Decimal('85.2')

        # Initial data ########################################################
        def request_callback(request):
            payload = {
                'ProductID': expected_id,
                'ProductName': expected_name,
                'Category': expected_category,
                'Price': float(expected_price),
            }

            resp_body = {'value': [payload]}
            headers = {}
            return requests.codes.ok, headers, json.dumps(resp_body)

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                rsps.GET, Product.__odata_url__(),
                callback=request_callback,
                content_type='application/json',
            )

            product = Service.query(Product).first()

        new_name = 'Changed value'

        # Patch call ##########################################################
        def request_callback_patch(request):
            payload = json.loads(request.body)
            assert 'ProductName' in payload
            assert payload['ProductName'] == new_name
            headers = {}
            return requests.codes.no_content, headers, ''

        # Reload call #########################################################
        def request_callback_reload(request):
            payload = {
                'ProductID': expected_id,
                'ProductName': new_name,
                'Category': expected_category,
                'Price': float(expected_price),
            }

            resp_body = {'value': [payload]}
            headers = {}
            return requests.codes.ok, headers, json.dumps(resp_body)

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                rsps.PATCH, product.__odata__.instance_url,
                callback=request_callback_patch,
                content_type='application/json',
            )
            rsps.add_callback(
                rsps.GET, product.__odata__.instance_url,
                callback=request_callback_reload,
                content_type='application/json',
            )

            product.name = new_name
            Service.save(product)

        assert product.name == new_name

    def test_delete(self):
        # Initial data ########################################################
        def request_callback(request):
            payload = {
                'ProductID': 2048,
                'ProductName': 'This product will be deleted',
                'Category': 'Something',
                'Price': 1234.5,
            }

            resp_body = {'value': [payload]}
            headers = {}
            return requests.codes.ok, headers, json.dumps(resp_body)

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                rsps.GET, Product.__odata_url__(),
                callback=request_callback,
                content_type='application/json',
            )

            product = Service.query(Product).first()

        # Delete call #########################################################
        def request_callback_delete(request):
            headers = {}
            return requests.codes.ok, headers, ''

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                rsps.DELETE, product.__odata__.instance_url,
                callback=request_callback_delete,
                content_type='application/json',
            )

            Service.delete(product)



# -*- coding: utf-8 -*-

from unittest import TestCase
import json

import responses
import requests

from odata.tests import Service, Product, ColorSelection


class TestEnums(TestCase):

    def test_insert_value(self):

        def request_callback(request):
            content = json.loads(request.body)
            content['ProductID'] = 1
            self.assertIn('ColorSelection', content)
            self.assertEqual(content.get('ColorSelection'), 'Black')
            headers = {}
            return requests.codes.ok, headers, json.dumps(content)

        with responses.RequestsMock() as rsps:
            rsps.add_callback(rsps.POST,
                              Product.__odata_url__(),
                              callback=request_callback,
                              content_type='application/json')

            new_product = Product()
            new_product.name = 'Test Product'
            new_product.color_selection = ColorSelection.Black
            Service.save(new_product)

    def test_read_value(self):
        test_product_values = dict(
            ProductID=1,
            ProductName='Test Product',
            Category='',
            ColorSelection='Red',
            Price=0.0,
        )
        with responses.RequestsMock() as rsps:
            rsps.add(rsps.GET, Product.__odata_url__(),
                     content_type='application/json',
                     json=dict(value=[test_product_values]))

            product = Service.query(Product).get(1)

        self.assertIsInstance(product.color_selection, ColorSelection)
        self.assertEqual(product.color_selection, ColorSelection.Red)



# -*- coding: utf-8 -*-

import unittest
import json

import responses
import requests

from odata.tests import Service, ProductWithNavigation, ProductPart, Manufacturer


class TestNavigationObjects(unittest.TestCase):

    def test_read_single_navigation_property(self):
        # Initial data ########################################################
        def request_callback(request):
            payload = {
                'ProductID': 51,
                'ProductName': 'Foo',
                'Category': 'Bar',
                'Price': 12.3,
            }

            resp_body = {'value': [payload]}
            headers = {}
            return requests.codes.ok, headers, json.dumps(resp_body)

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                rsps.GET, ProductWithNavigation.__odata_url__(),
                callback=request_callback,
                content_type='application/json',
            )

            product = Service.query(ProductWithNavigation).first()

        # Get parts ###########################################################
        parts_url = product.__odata__.instance_url + '/Manufacturer'

        def request_callback_manufacturer(request):
            payload = {
                'ManufacturerID': 33,
                'Name': 'Best Parts Ltd.',
                'DateEstablished': '2007-06-05T12:00:00Z',
            }

            resp_body = payload
            headers = {}
            return requests.codes.ok, headers, json.dumps(resp_body)

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                rsps.GET, parts_url,
                callback=request_callback_manufacturer,
                content_type='application/json',
            )
            mf = product.manufacturer

        assert isinstance(mf, Manufacturer)

    def test_read_collection_navigation_property(self):
        # Initial data ########################################################
        def request_callback(request):
            payload = {
                'ProductID': 51,
                'ProductName': 'Foo',
                'Category': 'Bar',
                'Price': 12.3,
            }

            resp_body = {'value': [payload]}
            headers = {}
            return requests.codes.ok, headers, json.dumps(resp_body)

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                rsps.GET, ProductWithNavigation.__odata_url__(),
                callback=request_callback,
                content_type='application/json',
            )
            product = Service.query(ProductWithNavigation).first()

        # Get parts ###########################################################
        parts_url = product.__odata__.instance_url + '/Parts'

        def request_callback_parts(request):
            payload = {
                'PartID': 512,
                'PartName': 'Bits and bobs',
                'Size': 5.333,
            }

            resp_body = {'value': [payload]}
            headers = {}
            return requests.codes.ok, headers, json.dumps(resp_body)

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                rsps.GET, parts_url,
                callback=request_callback_parts,
                content_type='application/json',
            )

            for part in product.parts:
                assert isinstance(part, ProductPart)

    def test_read_expanded_navigation_property(self):
        # Initial data ########################################################
        def request_callback(request):
            payload = {
                'ProductID': 51,
                'ProductName': 'Foo',
                'Category': 'Bar',
                'Price': 12.3,
                'Parts': [
                    {
                        'PartID': 512,
                        'PartName': 'Bits and bobs',
                        'Size': 5.333,
                    }
                ]
            }

            resp_body = {'value': [payload]}
            headers = {}
            return requests.codes.ok, headers, json.dumps(resp_body)

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                rsps.GET, ProductWithNavigation.__odata_url__(),
                callback=request_callback,
                content_type='application/json',
            )

            query = Service.query(ProductWithNavigation)
            query.expand(ProductWithNavigation.parts)
            product = query.first()

        # Get parts ###########################################################
        for part in product.parts:
            assert isinstance(part, ProductPart)

    def test_set_navigation_property(self):
        # Initial data ########################################################
        def request_callback(request):
            payload = {
                'ProductID': 51,
                'ProductName': 'Foo',
                'Category': 'Bar',
                'Price': 12.3,
            }

            resp_body = {'value': [payload]}
            headers = {}
            return requests.codes.ok, headers, json.dumps(resp_body)

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                rsps.GET, ProductWithNavigation.__odata_url__(),
                callback=request_callback,
                content_type='application/json',
            )

            product = Service.query(ProductWithNavigation).first()

        # Get part ############################################################
        def request_callback_parts(request):
            payload = {
                'PartID': 512,
                'PartName': 'Bits and bobs',
                'Size': 5.333,
            }

            resp_body = {'value': [payload]}
            headers = {}
            return requests.codes.ok, headers, json.dumps(resp_body)

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                rsps.GET, ProductPart.__odata_url__(),
                callback=request_callback_parts,
                content_type='application/json',
            )
            part = Service.query(ProductPart).first()

        product.parts = [part]

        # Patch call ##########################################################
        def request_callback_set_parts(request):
            payload = json.loads(request.body)

            key = 'Parts@odata.bind'
            assert key in payload
            assert payload.get(key) == ['ProductParts(512)']

            headers = {}
            return requests.codes.no_content, headers, ''

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                rsps.PATCH, product.__odata__.instance_url,
                callback=request_callback_set_parts,
                content_type='application/json',
            )

            # Reload data #####################################################
            rsps.add_callback(
                rsps.GET, product.__odata__.instance_url,
                callback=request_callback,
                content_type='application/json',
            )

            Service.save(product)



from ..odata.entity import EntityBase
from ..odata.entity import declarative_base
from ..odata.property import IntegerProperty, StringProperty, DatetimeProperty


class Phone(declarative_base()):
    __odata_type__ = 'ProductDataService.Objects.Product'
    __odata_collection__ = 'oData/Phones'

    Name = StringProperty('Name')
    IPAddress = StringProperty('IPAddress')
    DN = StringProperty('DN')
    Description = StringProperty('Description')
    Category = StringProperty('Category')
    User = StringProperty('User')
    DevicePool = StringProperty('DevicePool')
    Location = StringProperty('Location')
    Region = StringProperty('Region')
    CMGroup = StringProperty('CMGroup')
    DeviceCSS = StringProperty('DeviceCSS')
    PhoneLoad = StringProperty('PhoneLoad')
    ButtonTemplate = StringProperty('ButtonTemplate')
    AddonModule = StringProperty('AddonModule')
    EMEnabled = StringProperty('EMEnabled')
    LoginTime = StringProperty('LoginTime')
    Status = StringProperty('Status')
    ActiveServer = StringProperty('ActiveServer')
    Model = StringProperty('Model')
    LastUpdate = StringProperty('LastUpdate')
    LastRefresh = StringProperty('LastRefresh')
    Timestamp = StringProperty('Timestamp')
    LoginTime = StringProperty('LoginTime')
    Firmware = StringProperty('Firmware')
    ClusterGUID = StringProperty('ClusterGUID')
    DeviceState = StringProperty('DeviceState')
    LastActivity = StringProperty('LastActivity')

    def __str__(self):
        return "{0} ({1})".format(self.Name, self.DN)



# -*- coding: utf-8 -*-

import os
from unittest import TestCase
import json

import requests
import responses

from odata import ODataService
from odata.entity import EntityBase

path = os.path.join(os.path.dirname(__file__), 'demo_metadata.xml')
with open(path, mode='rb') as f:
    metadata_xml = f.read()


class TestMetadataImport(TestCase):

    def test_read(self):
        with responses.RequestsMock() as rsps:
            rsps.add(rsps.GET, 'http://demo.local/odata/$metadata/',
                     body=metadata_xml, content_type='text/xml')
            Service = ODataService('http://demo.local/odata/', reflect_entities=True)

        self.assertIn('Product', Service.entities)

        # non-entityset things should not be listed in entities
        expected_keys = {'Product', 'ProductWithNavigation', 'Manufacturer',
                         'ProductManufacturerSales'}
        self.assertEqual(set(Service.entities.keys()), expected_keys)

        Product = Service.entities['Product']
        ProductWithNavigation = Service.entities['ProductWithNavigation']

        assert issubclass(Product, EntityBase)
        assert hasattr(Product, 'DemoCollectionAction')

        test_product = Product()
        # shortcut for saving the entity
        test_product.__odata__.persisted = True
        assert hasattr(test_product, 'DemoActionWithParameters')
        assert hasattr(ProductWithNavigation, 'Manufacturer')
        self.assertIn('Manufacturer', Service.entities)

        self.assertIn('DemoUnboundAction', Service.actions)

    def test_computed_value_in_insert(self):
        with responses.RequestsMock() as rsps:
            rsps.add(rsps.GET, 'http://demo.local/odata/$metadata/',
                     body=metadata_xml, content_type='text/xml')
            Service = ODataService('http://demo.local/odata/', reflect_entities=True)

        Product = Service.entities['Product']
        test_product = Product()

        def request_callback_part(request):
            payload = json.loads(request.body)
            self.assertNotIn('ExampleComputed', payload)
            headers = {}
            return requests.codes.created, headers, json.dumps(payload)

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                rsps.POST, Product.__odata_url__(),
                callback=request_callback_part,
                content_type='application/json',
            )
            Service.save(test_product)



# -*- coding: utf-8 -*-

import unittest
import json

import requests
import responses

from odata.exceptions import ODataError
from odata.tests import Service, Product


class TestODataError(unittest.TestCase):

    def test_parse_error_json(self):
        expected_code = '0451'
        expected_message = 'Testing error message handling'
        expected_innererror_message = 'Detailed messages here'

        # Initial data ########################################################
        def request_callback(request):
            resp_body = {'error': {
                'code': expected_code,
                'message': expected_message,
                'innererror': {
                    'message': expected_innererror_message
                }
            }}
            headers = {
                'Content-Type': 'application/json;odata.metadata=minimal'
            }
            return requests.codes.bad_request, headers, json.dumps(resp_body)

        with responses.RequestsMock() as rsps:
            rsps.add_callback(
                rsps.GET, Product.__odata_url__(),
                callback=request_callback,
                content_type='application/json',
            )

            def action():
                try:
                    Service.query(Product).first()
                except ODataError as e:
                    errmsg = str(e)
                    assert expected_code in errmsg, 'Code not in text'
                    assert expected_message in errmsg, 'Upper level message not in text'
                    assert expected_innererror_message in errmsg, 'Detailed message not in text'
                    raise

            self.assertRaises(ODataError, action)



from setuptools import setup, find_packages

__version__ = '0.1.564'

setup(
    name='automationfx',
    packages=find_packages(),
    version=__version__,
    description='AutomationFX SDK for Cisco Unified Communication Manager',
    long_description=(
        "This is a SDK for the AutomationFX REST API."
        "AutomationFX is a integration platform from UnifiedFX that exposes Cisco Unified Communication Managers (CUCM) complex and varied interfaces via a simple unified REST API."
        "Note: Currently AutomationFX software is only availble via PhoneView Lab Editon (alpha release)."
        "Please contact UnifiedFX Sales for further information and request access to the software."
        ),
    author='Stephen Welsh',
    author_email='support@unifiedfx.com',
    url='https://github.com/unifiedfx/automationfx-python',
    download_url='https://github.com/unifiedfx/automationfx-python/archive/0.1.tar.gz',
    license='MIT',
    install_requires=['requests'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7'
        ],
    keywords='Cisco Phone Automation',
    zip_safe=False
    )


# -*- coding: utf-8 -*-

"""
Entity properties
=================

Entities are a collection of Properties of different types. These classes
implement the default set of types used in endpoints.

At entity class level, these properties are used in querying entities:

.. code-block:: python

    >>> import datetime
    >>> Order.ShippedDate > datetime.datetime.now()
    'ShippedDate gt 2016-02-19T12:02:04.956226'
    >>> Service.query(Order).filter(Order.OrderID == 1234)

Once the entity is instanced, the properties act as data getters and setters:

.. code-block:: python

    >>> order = Order()
    >>> order.ShippedDate = datetime.datetime.now()
    >>> Service.save(order)

Setting a new value to a property marks the property as `dirty`, and will be
sent to the endpoint when :py:func:`~odata.service.ODataService.save` is
called.

This behavior is similar to SQLAlchemy's ORM.


.. automodule:: odata.navproperty


Creating new property types
---------------------------

All properties must subclass :py:class:`PropertyBase`, and implement the
serialization methods. You can then use the created Property class in your
custom defined entities. Replacing the default types is not supported.

.. autoclass:: odata.property.PropertyBase
    :members:

----

Types
-----
"""

from decimal import Decimal
import datetime

import dateutil.parser

from .navproperty import NavigationProperty


class PropertyBase(object):
    """
    A base class for all properties.

    :param name: Name of the property in the endpoint
    :param primary_key: This property is a primary key
    :param is_collection: This property contains multiple values
    """
    def __init__(self, name, primary_key=False, is_collection=False, is_computed_value=False):
        """
        :type name: str
        :type primary_key: bool
        """
        self.name = name
        self.primary_key = primary_key
        self.is_collection = is_collection
        self.is_computed_value = is_computed_value

    def __repr__(self):
        return '<Property({0})>'.format(self.name)

    def __get__(self, instance, owner):
        """
        :type instance: odata.entity.EntityBase
        :type owner: odata.entity.EntityBase
        """
        if instance is None:
            return self

        es = instance.__odata__

        if self.name in es:
            raw_data = es[self.name]
            if self.is_collection:
                if raw_data is None:
                    return

                data = []
                for i in raw_data:
                    data.append(self.deserialize(i))
                return data
            else:
                return self.deserialize(raw_data)
        else:
            raise AttributeError()

    def __set__(self, instance, value):
        """
        :type instance: odata.entity.EntityBase
        """

        es = instance.__odata__

        if self.name in es:
            if self.is_collection:
                data = []
                for i in (value or []):
                    data.append(self.serialize(i))
                new_value = data
            else:
                new_value = self.serialize(value)
            old_value = es[self.name]
            if new_value != old_value:
                es[self.name] = new_value
                es.set_property_dirty(self)

    def serialize(self, value):
        """
        Called when serializing the value to JSON. Implement this method when
        creating a new Property class

        :param value: Value given in Python code
        :returns: Value that will be used in JSON
        """
        raise NotImplementedError()

    def deserialize(self, value):
        """
        Called when deserializing the value from JSON to Python. Implement this
        method when creating a new Property class

        :param value: Value received in JSON
        :returns: Value that will be passed to Python
        """
        raise NotImplementedError()

    def escape_value(self, value):
        """
        Called when escaping the property value for usage in Query string.
        Implement this method when creating a new Property class

        :param value: Value of this property
        :return: Escaped value that can be used in Query string
        """
        if value is None:
            return 'null'
        return value

    def asc(self):
        return '{0} asc'.format(self.name)

    def desc(self):
        return '{0} desc'.format(self.name)

    def __eq__(self, other):
        value = self.escape_value(other)
        return u'{0} eq {1}'.format(self.name, value)

    def __ne__(self, other):
        value = self.escape_value(other)
        return u'{0} ne {1}'.format(self.name, value)

    def __ge__(self, other):
        value = self.escape_value(other)
        return u'{0} ge {1}'.format(self.name, value)

    def __gt__(self, other):
        value = self.escape_value(other)
        return u'{0} gt {1}'.format(self.name, value)

    def __le__(self, other):
        value = self.escape_value(other)
        return u'{0} le {1}'.format(self.name, value)

    def __lt__(self, other):
        value = self.escape_value(other)
        return u'{0} lt {1}'.format(self.name, value)

    def startswith(self, value):
        value = self.escape_value(value)
        return u'startswith({0}, {1})'.format(self.name, value)

    def endswith(self, value):
        value = self.escape_value(value)
        return u'endswith({0}, {1})'.format(self.name, value)

    def contains(self, value):
        value = self.escape_value(value)
        return u'contains({0}, {1})'.format(self.name, value)


class IntegerProperty(PropertyBase):
    """
    Property that stores a plain old integer
    """
    def serialize(self, value):
        return value

    def deserialize(self, value):
        return value


class StringProperty(PropertyBase):
    """
    Property that stores a unicode string
    """
    def serialize(self, value):
        return value

    def deserialize(self, value):
        return value

    def escape_value(self, value):
        if value is None:
            return 'null'
        return u"'{0}'".format(value.replace("'", "''"))


class BooleanProperty(PropertyBase):
    """
    Property that stores a boolean value
    """
    def escape_value(self, value):
        if value:
            return 'true'
        return 'false'

    def serialize(self, value):
        return bool(value)

    def deserialize(self, value):
        return bool(value)


class FloatProperty(PropertyBase):
    """
    Property that stores a float value
    """
    def serialize(self, value):
        return value

    def deserialize(self, value):
        return value


class DecimalProperty(PropertyBase):
    """
    Property that stores a decimal value. JSON does not support this directly,
    so the value will be transmitted as a float
    """
    def escape_value(self, value):
        if value is None:
            return 'null'
        return str(value)

    def serialize(self, value):
        if value is not None:
            return float(value)

    def deserialize(self, value):
        if value is not None:
            return Decimal(str(value))


class DatetimeProperty(PropertyBase):
    """
    Property that stores a datetime object. JSON does not support date objects
    natively so dates are transmitted as ISO-8601 formatted strings
    """
    def escape_value(self, value):
        if value is None:
            return 'null'
        return value.isoformat()

    def serialize(self, value):
        if isinstance(value, datetime.datetime):
            r = value.isoformat()
            if value.tzinfo is None:
                r += 'Z'
            return r

    def deserialize(self, value):
        if value:
            return dateutil.parser.parse(value)


class UUIDProperty(StringProperty):
    """
    Property that stores a UUID (also known as GUID) value. JSON does not
    support this directly, so the value will be transmitted as a string.
    Unlike :py:class:`StringProperty`, it does not escape quotes as query
    filters do not use quotes for UUID
    """
    def serialize(self, value):
        return str(value)

    def deserialize(self, value):
        return str(value)

    def escape_value(self, value):
        if value is None:
            return 'null'
        return str(value)



# -*- coding: utf-8 -*-

import logging
import sys
has_lxml = False
try:
    from lxml import etree as ET
    has_lxml = True
except ImportError:
    if sys.version_info < (2, 7):
        raise ImportError('lxml required for Python versions older than 2.7')
    from xml.etree import ElementTree as ET

from .entity import declarative_base, EntityBase
from .exceptions import ODataReflectionError
from .property import StringProperty, IntegerProperty, DecimalProperty, \
    DatetimeProperty, BooleanProperty, NavigationProperty, UUIDProperty
from .enumtype import EnumType, EnumTypeProperty


class MetaData(object):

    log = logging.getLogger('odata.metadata')
    namespaces = {
        'edm': 'http://docs.oasis-open.org/odata/ns/edm',
        'edmx': 'http://docs.oasis-open.org/odata/ns/edmx'
    }

    property_types = {
        'Edm.Int16': IntegerProperty,
        'Edm.Int32': IntegerProperty,
        'Edm.Int64': IntegerProperty,
        'Edm.String': StringProperty,
        'Edm.Single': DecimalProperty,
        'Edm.Decimal': DecimalProperty,
        'Edm.DateTimeOffset': DatetimeProperty,
        'Edm.Boolean': BooleanProperty,
        'Edm.Guid': UUIDProperty,
    }

    _annotation_term_computed = 'Org.OData.Core.V1.Computed'

    def __init__(self, service):
        self.url = service.url + '$metadata/'
        self.connection = service.default_context.connection
        self.service = service

    def property_type_to_python(self, edm_type):
        return self.property_types.get(edm_type, StringProperty)

    def _type_is_collection(self, typename):
        if typename.startswith('Collection('):
            stripped = typename[11:-1]
            return True, stripped
        else:
            return False, typename

    def _set_object_relationships(self, entities):
        for entity in entities.values():
            schema = entity.__odata_schema__

            for schema_nav in schema.get('navigation_properties', []):
                name = schema_nav['name']
                type_ = schema_nav['type']
                foreign_key = schema_nav['foreign_key']

                is_collection, type_ = self._type_is_collection(type_)

                for _search_entity in entities.values():
                    _search_type = _search_entity.__odata_schema__['type']
                    _search_type_alias = _search_entity.__odata_schema__.get('type_alias')
                    if type_ in (_search_type, _search_type_alias):
                        nav = NavigationProperty(
                            name,
                            _search_entity,
                            collection=is_collection,
                            foreign_key=foreign_key,
                        )
                        setattr(entity, name, nav)

    def _create_entities(self, all_types, entities, entity_sets, entity_base_class, schemas, depth=1):
        orphan_entities = []
        for schema in schemas:
            for entity_dict in schema.get('entities'):
                entity_type = entity_dict['type']
                entity_type_alias = entity_dict.get('type_alias')
                entity_name = entity_dict['name']

                if entity_type in all_types:
                    continue

                if entity_dict.get('base_type'):
                    base_type = entity_dict.get('base_type')
                    parent_entity_class = all_types.get(base_type)

                    if parent_entity_class is None:
                        # base class not yet created
                        orphan_entities.append(entity_type)
                        continue

                    object_dict = dict(
                        __odata_schema__=entity_dict,
                        __odata_type__=entity_type,
                    )
                    entity_set = entity_sets.get(entity_type) or entity_sets.get(entity_type_alias)
                    collection_name = (entity_set or {}).get('name')
                    if collection_name:
                        object_dict['__odata_collection__'] = collection_name

                    entity_class = type(entity_name, (parent_entity_class,), object_dict)  # type: EntityBase
                    if entity_class.__odata_collection__:
                        entities[entity_name] = entity_class
                else:
                    entity_set = entity_sets.get(entity_type) or entity_sets.get(entity_type_alias)
                    collection_name = (entity_set or {}).get('name')
                    object_dict = dict(
                        __odata_schema__=entity_dict,
                        __odata_type__=entity_type,
                        __odata_collection__=collection_name
                    )
                    entity_class = type(entity_name, (entity_base_class,), object_dict)
                    if collection_name:
                        entities[entity_name] = entity_class

                all_types[entity_type] = entity_class
                if entity_type_alias:
                    all_types[entity_type_alias] = entity_class

                for prop in entity_dict.get('properties'):
                    prop_name = prop['name']

                    if hasattr(entity_class, prop_name):
                        # do not replace existing properties (from Base)
                        continue

                    property_type = all_types.get(prop['type'])

                    if property_type and issubclass(property_type, EnumType):
                        property_instance = EnumTypeProperty(prop_name, enum_class=property_type)
                        property_instance.is_computed_value = prop['is_computed_value']
                    else:
                        type_ = self.property_type_to_python(prop['type'])
                        type_options = {
                            'primary_key': prop['is_primary_key'],
                            'is_collection': prop['is_collection'],
                            'is_computed_value': prop['is_computed_value'],
                        }
                        property_instance = type_(prop_name, **type_options)
                    setattr(entity_class, prop_name, property_instance)

        if len(orphan_entities) > 0:
            if depth > 10:
                errmsg = ('Types could not be resolved. '
                          'Orphaned types: {0}').format(', '.join(orphan_entities))
                raise ODataReflectionError(errmsg)
            depth += 1
            self._create_entities(all_types, entities, entity_sets, entity_base_class, schemas, depth)

    def _create_actions(self, entities, actions, get_entity_or_prop_from_type):
        for action in actions:
            entity_type = action['is_bound_to']
            bind_entity = None
            bound_to_collection = False
            if entity_type:
                bound_to_collection, entity_type = self._type_is_collection(entity_type)
                for entity in entities.values():
                    schema = entity.__odata_schema__
                    if entity_type in (schema['type'], schema.get('type_alias')):
                        bind_entity = entity

            parameters_dict = {}
            for param in action['parameters']:
                parameters_dict[param['name']] = self.property_type_to_python(param['type'])

            object_dict = dict(
                __odata_service__=self.service,
                name=action['fully_qualified_name'],
                parameters=parameters_dict,
                return_type=get_entity_or_prop_from_type(action['return_type']),
                return_type_collection=get_entity_or_prop_from_type(action['return_type_collection']),
                bound_to_collection=bound_to_collection,
            )
            action_class = type(action['name'], (self.service.Action,), object_dict)

            if bind_entity:
                setattr(bind_entity, action['name'], action_class())
            else:
                self.service.actions[action['name']] = action_class()

    def _create_functions(self, entities, functions, get_entity_or_prop_from_type):
        for function in functions:
            entity_type = function['is_bound_to']
            bind_entity = None
            bound_to_collection = False
            if entity_type:
                bound_to_collection, entity_type = self._type_is_collection(entity_type)
                for entity in entities.values():
                    schema = entity.__odata_schema__
                    if entity_type in (schema['type'], schema.get('type_alias')):
                        bind_entity = entity

            parameters_dict = {}
            for param in function['parameters']:
                parameters_dict[param['name']] = self.property_type_to_python(param['type'])

            object_dict = dict(
                __odata_service__=self.service,
                name=function['fully_qualified_name'],
                parameters=parameters_dict,
                return_type=get_entity_or_prop_from_type(function['return_type']),
                return_type_collection=get_entity_or_prop_from_type(function['return_type_collection']),
                bound_to_collection=bound_to_collection,
            )
            function_class = type(function['name'], (self.service.Function,), object_dict)

            if bind_entity:
                setattr(bind_entity, function['name'], function_class())
            else:
                self.service.functions[function['name']] = function_class()

    def get_entity_sets(self, base=None):
        document = self.load_document()
        schemas, entity_sets, actions, functions = self.parse_document(document)

        entities = {}
        base_class = base or declarative_base()
        all_types = {}

        def get_entity_or_prop_from_type(typename):
            if typename is None:
                return

            type_ = all_types.get(typename)
            if type_ is not None:
                return type_

            return self.property_type_to_python(typename)

        for schema in schemas:
            for enum_type in schema['enum_types']:
                names = [(i['name'], i['value']) for i in enum_type['members']]
                created_enum = EnumType(enum_type['name'], names=names)
                all_types[enum_type['fully_qualified_name']] = created_enum

        self._create_entities(all_types, entities, entity_sets, base_class, schemas)
        self._set_object_relationships(entities)
        self._create_actions(entities, actions, get_entity_or_prop_from_type)
        self._create_functions(entities, functions, get_entity_or_prop_from_type)

        self.log.info('Loaded {0} entity sets, total {1} types'.format(len(entities), len(all_types)))
        return base_class, entities, all_types

    def load_document(self):
        self.log.info('Loading metadata document: {0}'.format(self.url))
        response = self.connection._do_get(self.url)
        return ET.fromstring(response.content)

    def _parse_action(self, xmlq, action_element, schema_name):
        action = {
            'name': action_element.attrib['Name'],
            'fully_qualified_name': action_element.attrib['Name'],
            'is_bound': action_element.attrib.get('IsBound') == 'true',
            'is_bound_to': None,
            'parameters': [],
            'return_type': None,
            'return_type_collection': None,
        }

        if action['is_bound']:
            # bound actions are named SchemaNamespace.ActionName
            action['fully_qualified_name'] = '.'.join([schema_name, action['name']])

        for parameter_element in xmlq(action_element, 'edm:Parameter'):
            parameter_name = parameter_element.attrib['Name']
            if action['is_bound'] and parameter_name == 'bindingParameter':
                action['is_bound_to'] = parameter_element.attrib['Type']
                continue

            parameter_type = parameter_element.attrib['Type']

            action['parameters'].append({
                'name': parameter_name,
                'type': parameter_type,
            })

        for return_type_element in xmlq(action_element, 'edm:ReturnType'):
            type_name = return_type_element.attrib['Type']
            is_collection, type_name = self._type_is_collection(type_name)
            if is_collection:
                action['return_type_collection'] = type_name
            else:
                action['return_type'] = type_name
        return action

    def _parse_function(self, xmlq, function_element, schema_name):
        function = {
            'name': function_element.attrib['Name'],
            'fully_qualified_name': function_element.attrib['Name'],
            'is_bound': function_element.attrib.get('IsBound') == 'true',
            'is_bound_to': None,
            'parameters': [],
            'return_type': None,
            'return_type_collection': None,
        }

        if function['is_bound']:
            # bound functions are named SchemaNamespace.FunctionName
            function['fully_qualified_name'] = '.'.join(
                [schema_name, function['name']])

        for parameter_element in xmlq(function_element, 'edm:Parameter'):
            parameter_name = parameter_element.attrib['Name']
            if function['is_bound'] and parameter_name == 'bindingParameter':
                function['is_bound_to'] = parameter_element.attrib['Type']
                continue

            parameter_type = parameter_element.attrib['Type']

            function['parameters'].append({
                'name': parameter_name,
                'type': parameter_type,
            })

        for return_type_element in xmlq(function_element, 'edm:ReturnType'):
            type_name = return_type_element.attrib['Type']
            is_collection, type_name = self._type_is_collection(type_name)
            if is_collection:
                function['return_type_collection'] = type_name
            else:
                function['return_type'] = type_name
        return function

    def _parse_entity(self, xmlq, entity_element, schema_name, schema_alias):
        entity_name = entity_element.attrib['Name']

        entity_type_name = '.'.join([schema_name, entity_name])

        entity = {
            'name': entity_name,
            'type': entity_type_name,
            'properties': [],
            'navigation_properties': [],
        }

        if schema_alias:
            entity_type_name_alias = '.'.join([schema_alias, entity_name])
            entity['type_alias'] = entity_type_name_alias

        base_type = entity_element.attrib.get('BaseType')
        if base_type:
            entity['base_type'] = base_type

        entity_pks = {}
        for pk_property in xmlq(entity_element, 'edm:Key/edm:PropertyRef'):
            pk_property_name = pk_property.attrib['Name']
            entity_pks[pk_property_name] = 0

        for entity_property in xmlq(entity_element, 'edm:Property'):
            p_name = entity_property.attrib['Name']
            p_type = entity_property.attrib['Type']

            is_collection, p_type = self._type_is_collection(p_type)
            is_computed_value = False

            for annotation in xmlq(entity_property, 'edm:Annotation'):
                annotation_term = annotation.attrib.get('Term', '')
                annotation_bool = annotation.attrib.get('Bool') == 'true'
                if annotation_term == self._annotation_term_computed:
                    is_computed_value = annotation_bool

            entity['properties'].append({
                'name': p_name,
                'type': p_type,
                'is_primary_key': p_name in entity_pks,
                'is_collection': is_collection,
                'is_computed_value': is_computed_value,
            })

        for nav_property in xmlq(entity_element, 'edm:NavigationProperty'):
            p_name = nav_property.attrib['Name']
            p_type = nav_property.attrib['Type']
            p_foreign_key = None

            ref_constraint = xmlq(nav_property, 'edm:ReferentialConstraint')
            if ref_constraint:
                ref_constraint = ref_constraint[0]
                p_foreign_key = ref_constraint.attrib['Property']

            entity['navigation_properties'].append({
                'name': p_name,
                'type': p_type,
                'foreign_key': p_foreign_key,
            })
        return entity

    def _parse_enumtype(self, xmlq, enumtype_element, schema_name):
        enum_name = enumtype_element.attrib['Name']
        enum = {
            'name': enum_name,
            'fully_qualified_name': '.'.join([schema_name, enum_name]),
            'members': []
        }
        for enum_member in xmlq(enumtype_element, 'edm:Member'):
            member_name = enum_member.attrib['Name']
            member_value = int(enum_member.attrib['Value'])
            enum['members'].append({
                'name': member_name,
                'value': member_value,
            })
        return enum

    def parse_document(self, doc):
        schemas = []
        container_sets = {}
        actions = []
        functions = []

        if has_lxml:
            def xmlq(node, xpath):
                return node.xpath(xpath, namespaces=self.namespaces)
        else:
            def xmlq(node, xpath):
                return node.findall(xpath, namespaces=self.namespaces)

        for schema in xmlq(doc, 'edmx:DataServices/edm:Schema'):
            schema_name = schema.attrib['Namespace']
            schema_alias = schema.attrib.get('Alias')

            schema_dict = {
                'name': schema_name,
                'alias': schema_alias,
                'entities': [],
                'enum_types': [],
                'complex_types': [],
            }

            for enum_type in xmlq(schema, 'edm:EnumType'):
                enum = self._parse_enumtype(xmlq, enum_type, schema_name)
                schema_dict['enum_types'].append(enum)

            for entity_type in xmlq(schema, 'edm:EntityType'):
                entity = self._parse_entity(xmlq, entity_type, schema_name, schema_alias)
                schema_dict['entities'].append(entity)

            schemas.append(schema_dict)

        for schema in xmlq(doc, 'edmx:DataServices/edm:Schema'):
            schema_name = schema.attrib['Namespace']
            for entity_set in xmlq(schema, 'edm:EntityContainer/edm:EntitySet'):
                set_name = entity_set.attrib['Name']
                set_type = entity_set.attrib['EntityType']

                set_dict = {
                    'name': set_name,
                    'type': set_type,
                    'schema': None,
                }

                for schema_ in schemas:
                    for entity in schema_.get('entities', []):
                        if set_type in (entity.get('type'), entity.get('type_alias')):
                            set_dict['schema'] = entity

                container_sets[set_type] = set_dict

            for action_def in xmlq(schema, 'edm:Action'):
                action = self._parse_action(xmlq, action_def, schema_name)
                actions.append(action)

            for function_def in xmlq(schema, 'edm:Function'):
                function = self._parse_function(xmlq, function_def, schema_name)
                functions.append(function)

        return schemas, container_sets, actions, functions
